/*
	FastEDIUpsizer plugin for Avisynth v2.5.x.  FastEDIUpsizer upscales video
	by a factor of two using NEDI with a constant 8x8 window size on luma and
	bicubic interpolation on the chroma planes.  Only works in YV12 colorspace.

	Copyright (C) 2004 Kevin Stone

    v0.9 - main code

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "FastEDIUpsizer.h"

PVideoFrame __stdcall FastEDIUpsizer::GetFrame(int frameNum, IScriptEnvironment* env)
{
	if (frameNum < 0) frameNum = 0;
	else if (frameNum > nfrms) frameNum = nfrms;
	PVideoFrame Src = child->GetFrame(frameNum, env);
	PVideoFrame Dst = env->NewVideoFrame(vi);
	const unsigned char *srcpY = Src->GetReadPtr(PLANAR_Y);
	const unsigned char *srcpV = Src->GetReadPtr(PLANAR_V);
	const unsigned char *srcpU = Src->GetReadPtr(PLANAR_U);
	int src_pitchY = Src->GetPitch(PLANAR_Y);
	int src_pitchUV = Src->GetPitch(PLANAR_V);
	int src_widthY = Src->GetRowSize(PLANAR_Y);
	int src_widthUV = Src->GetRowSize(PLANAR_V);
	int src_heightY = Src->GetHeight(PLANAR_Y);
	int src_heightUV = Src->GetHeight(PLANAR_V);
	unsigned char *dstpY = Dst->GetWritePtr(PLANAR_Y);
	unsigned char *dstpV = Dst->GetWritePtr(PLANAR_V);
	unsigned char *dstpU = Dst->GetWritePtr(PLANAR_U);
	int dst_pitchY = Dst->GetPitch(PLANAR_Y);
	int dst_pitchY2 = dst_pitchY<<1;
	int dst_pitchY3 = dst_pitchY2 + dst_pitchY;
	int dst_pitchUV = Dst->GetPitch(PLANAR_V);
	int dst_pitchUV2 = dst_pitchUV<<1;
	int dst_widthY = Dst->GetRowSize(PLANAR_Y);
	int dst_widthUV = Dst->GetRowSize(PLANAR_V);
	int dst_heightY = Dst->GetHeight(PLANAR_Y);
	int dst_heightUV = Dst->GetHeight(PLANAR_V);
	unsigned char *dY = dstpY, *dV = dstpV, *dU = dstpU;
	int x, y, u, v, temp, temp1, temp2;
	/* Copy over original pixels */
	for (y=0; y<src_heightY; ++y)
	{
		for (x=0; x<src_widthY; ++x)
			dstpY[x<<1] = srcpY[x];
		dstpY += dst_pitchY2;
		srcpY += src_pitchY;
	}
	for (y=0; y<src_heightUV; ++y)
	{
		for (x=0; x<src_widthUV; ++x)
		{
			dstpV[x<<1] = srcpV[x];
			dstpU[x<<1] = srcpU[x];
		}
		dstpV += dst_pitchUV2;
		dstpU += dst_pitchUV2;
		srcpV += src_pitchUV;
		srcpU += src_pitchUV;
	}
	/* Interpolate areas in luma where NEDI can't be used */
	dstpY = dY;
	temp2 = dst_widthY-4;
	for (y=0; y<11; y+=2)
	{
		dstpY[1] = (dstpY[0] + dstpY[2] + 1) >> 1;
		for (x=3; x<temp2; x+=2) 
			dstpY[x] = (*this.*cPtr)(dstpY[x-3],dstpY[x-1],dstpY[x+1],dstpY[x+3]);
		dstpY[x] = (dstpY[x-1] + dstpY[x+1] + 1) >> 1; x+=2;
		dstpY[x] = dstpY[x-1];
		dstpY += dst_pitchY2;
	}
	temp1 = dst_heightY-12;
	temp2 = dst_widthY-4;
	for (; y<temp1; y+=2)
	{
		dstpY[1] = (dstpY[0] + dstpY[2] + 1) >> 1;
		for (x=3; x<9; x+=2) 
			dstpY[x] = (*this.*cPtr)(dstpY[x-3],dstpY[x-1],dstpY[x+1],dstpY[x+3]);
		for (x=dst_widthY-9; x<temp2; x+=2) 
			dstpY[x] = (*this.*cPtr)(dstpY[x-3],dstpY[x-1],dstpY[x+1],dstpY[x+3]);
		dstpY[x] = (dstpY[x-1] + dstpY[x+1] + 1) >> 1; x+=2;
		dstpY[x] = dstpY[x-1];
		dstpY += dst_pitchY2;
	}
	temp2 = dst_widthY-4;
	for (; y<dst_heightY; y+=2)
	{
		dstpY[1] = (dstpY[0] + dstpY[2] + 1) >> 1;
		for (x=3; x<temp2; x+=2) 
			dstpY[x] = (*this.*cPtr)(dstpY[x-3],dstpY[x-1],dstpY[x+1],dstpY[x+3]);
		dstpY[x] = (dstpY[x-1] + dstpY[x+1] + 1) >> 1; x+=2;
		dstpY[x] = dstpY[x-1];
		dstpY += dst_pitchY2;
	}
	dstpY = dY + dst_pitchY;
	unsigned char *dstppY = dstpY - dst_pitchY;
	unsigned char *dstpppY = dstppY - dst_pitchY2;
	unsigned char *dstpnY = dstpY + dst_pitchY;
	unsigned char *dstpnnY = dstpnY + dst_pitchY2;
	for (x=0; x<dst_widthY; ++x) 
		dstpY[x] = (dstppY[x] + dstpnY[x] + 1) >> 1;
	dstpppY = dstppY;
	dstppY = dstpnY;
	dstpY += dst_pitchY2;
	dstpnY = dstpnnY;
	dstpnnY += dst_pitchY2;
	for (y=3; y<9; y+=2)
	{
		for (x=0; x<dst_widthY; ++x) 
			dstpY[x] = (*this.*cPtr)(dstpppY[x],dstppY[x],dstpnY[x],dstpnnY[x]);
		dstpppY = dstppY;
		dstppY = dstpnY;
		dstpY += dst_pitchY2;
		dstpnY = dstpnnY;
		dstpnnY += dst_pitchY2;
	}
	temp1 = dst_heightY-10;
	for (; y<temp1; y+=2)
	{
		for (x=0; x<9; ++x) 
			dstpY[x] = (*this.*cPtr)(dstpppY[x],dstppY[x],dstpnY[x],dstpnnY[x]);
		for (x=dst_widthY-9; x<dst_widthY; ++x) 
			dstpY[x] = (*this.*cPtr)(dstpppY[x],dstppY[x],dstpnY[x],dstpnnY[x]);
		dstpppY = dstppY;
		dstppY = dstpnY;
		dstpY += dst_pitchY2;
		dstpnY = dstpnnY;
		dstpnnY += dst_pitchY2;
	}
	temp1 = dst_heightY-4;
	for (; y<temp1; y+=2)
	{
		for (x=0; x<dst_widthY; ++x) 
			dstpY[x] = (*this.*cPtr)(dstpppY[x],dstppY[x],dstpnY[x],dstpnnY[x]);
		dstpppY = dstppY;
		dstppY = dstpnY;
		dstpY += dst_pitchY2;
		dstpnY = dstpnnY;
		dstpnnY += dst_pitchY2;
	}
	for (x=0; x<dst_widthY; ++x) 
		dstpY[x] = (dstppY[x] + dstpnY[x] + 1) >> 1;
	dstppY = dstpnY;
	dstpY += dst_pitchY2;
	for (x=0; x<dst_widthY; ++x) dstpY[x] = dstppY[x];
	/* NEDI -- (luma only) */
	localPat1[0] = -dst_pitchY2-2; localPat1[1] = -dst_pitchY2+2; 
	localPat1[2] = dst_pitchY2-2; localPat1[3] = dst_pitchY2+2;
	localPat2[0] = -2; localPat2[1] = -dst_pitchY2; 
	localPat2[2] = dst_pitchY2; localPat2[3] = 2;
	windowPat161[0] = -dst_pitchY3-3; windowPat161[1] = -dst_pitchY3-1;   
	windowPat161[2] = -dst_pitchY-3; windowPat161[3] = -dst_pitchY-1;
	windowPat161[4] = -dst_pitchY3+1; windowPat161[5] = -dst_pitchY3+3;   
	windowPat161[6] = -dst_pitchY+1; windowPat161[7] = -dst_pitchY+3;
	windowPat161[8] = dst_pitchY-3; windowPat161[9] = dst_pitchY-1;
	windowPat161[10] = dst_pitchY3-3; windowPat161[11] = dst_pitchY3-1;
	windowPat161[12] = dst_pitchY+1; windowPat161[13] = dst_pitchY+3;
	windowPat161[14] = dst_pitchY3+1; windowPat161[15] = dst_pitchY3+3;
	windowPat162[0] = -3; windowPat162[1] = -dst_pitchY-2; 
	windowPat162[2] = dst_pitchY-2; windowPat162[3] = -1;
	windowPat162[4] = -dst_pitchY2-1; windowPat162[5] = -dst_pitchY3; 
	windowPat162[6] = -dst_pitchY; windowPat162[7] = -dst_pitchY2+1;
	windowPat162[8] = dst_pitchY2-1; windowPat162[9] = dst_pitchY;
	windowPat162[10] = dst_pitchY3; windowPat162[11] = dst_pitchY2+1;
	windowPat162[12] = 1; windowPat162[13] = -dst_pitchY+2; 
	windowPat162[14] = dst_pitchY+2; windowPat162[15] = 3;
	for(v=0; v<4; ++v)
	{
		for(u=0; u<16; ++u)
		{
			windowPat641[(v<<4)+u] = windowPat161[u] + (localPat1[v]<<1);
			windowPat642[(v<<4)+u] = windowPat162[u] + (localPat2[v]<<1);
		}
	}
	dstpY = dY + 9*dst_pitchY;
	temp1 = dst_heightY-9;
	temp2 = dst_widthY-9;
	for (y=9; y<temp1; y+=2)
	{
		for (x=9; x<temp2; x+=2)
		{
			pc[0] = dstpY[x+windowPat641[15]];
			pc[1] = dstpY[x+windowPat641[26]];
			pc[2] = dstpY[x+windowPat641[37]];
			pc[3] = dstpY[x+windowPat641[48]];
			if (checkVariance(pc))
			{
				for(u=0; u<64; ++u)
				{
					p[u] = dstpY[x + windowPat641[u]];
					n[u][0] = dstpY[x + windowPat641[u] + localPat1[0]];
					n[u][1] = dstpY[x + windowPat641[u] + localPat1[1]];
					n[u][2] = dstpY[x + windowPat641[u] + localPat1[2]];
					n[u][3] = dstpY[x + windowPat641[u] + localPat1[3]];
				}
				dstpY[x] = nEDI_8x8(p, n, 1);
			}
			else
			{
				unsigned char temp1 = (*this.*cPtr)(dstpY[x+windowPat641[12]],dstpY[x+windowPat641[13]],
													dstpY[x+windowPat641[24]],dstpY[x+windowPat641[25]]);
				unsigned char temp2 = (*this.*cPtr)(dstpY[x+windowPat641[14]],dstpY[x+windowPat641[15]],
													dstpY[x+windowPat641[26]],dstpY[x+windowPat641[27]]);
				unsigned char temp3 = (*this.*cPtr)(dstpY[x+windowPat641[36]],dstpY[x+windowPat641[37]],
													dstpY[x+windowPat641[48]],dstpY[x+windowPat641[49]]);
				unsigned char temp4 = (*this.*cPtr)(dstpY[x+windowPat641[38]],dstpY[x+windowPat641[39]],
													dstpY[x+windowPat641[50]],dstpY[x+windowPat641[51]]);
				dstpY[x] = (*this.*cPtr)(temp1,temp2,temp3,temp4);
			}
		}
		dstpY += dst_pitchY2;
	}
	dstpY = dY + 9*dst_pitchY;
	for (y=9; y<temp1; ++y)
	{
		temp = y&1;
		for (x=9+temp; x<temp2; x+=2)
		{
			pc[0] = dstpY[x+windowPat642[15]];
			pc[1] = dstpY[x+windowPat642[26]];
			pc[2] = dstpY[x+windowPat642[37]];
			pc[3] = dstpY[x+windowPat642[48]];
			if (checkVariance(pc))
			{
				for(u=0; u<64; ++u)
				{
					p[u] = dstpY[x + windowPat642[u]];
					n[u][0] = dstpY[x + windowPat642[u] + localPat2[0]];
					n[u][1] = dstpY[x + windowPat642[u] + localPat2[1]];
					n[u][2] = dstpY[x + windowPat642[u] + localPat2[2]];
					n[u][3] = dstpY[x + windowPat642[u] + localPat2[3]];
				}
				dstpY[x] = nEDI_8x8(p, n, 2+temp);
			}
			else
			{
				if (temp)
					dstpY[x] = (*this.*cPtr)(dstpY[x+windowPat642[25]],dstpY[x+windowPat642[26]],
									dstpY[x+windowPat642[37]],dstpY[x+windowPat642[38]]);
				else
					dstpY[x] = (*this.*cPtr)(dstpY[x-3],dstpY[x-1],dstpY[x+1],dstpY[x+3]);
			}
		}
		dstpY += dst_pitchY;
	}
	/* Interpolate Chroma */
	dstpV = dV; dstpU = dU;
	temp2 = dst_widthUV-4;
	for (y=0; y<dst_heightUV; y+=2)
	{
		dstpV[1] = (dstpV[0] + dstpV[2] + 1) >> 1;
		dstpU[1] = (dstpU[0] + dstpU[2] + 1) >> 1;
		for (x=3; x<temp2; x+=2)
		{
			dstpV[x] = (*this.*cPtr)(dstpV[x-3],dstpV[x-1],dstpV[x+1],dstpV[x+3]);
			dstpU[x] = (*this.*cPtr)(dstpU[x-3],dstpU[x-1],dstpU[x+1],dstpU[x+3]);
		}
		dstpV[x] = (dstpV[x-1] + dstpV[x+1] + 1) >> 1;
		dstpU[x] = (dstpU[x-1] + dstpU[x+1] + 1) >> 1; x+=2;
		dstpV[x] = dstpV[x-1];
		dstpU[x] = dstpU[x-1];
		dstpV += dst_pitchUV2;
		dstpU += dst_pitchUV2;
	}
	dstpV = dV + dst_pitchUV; dstpU = dU + dst_pitchUV;
	unsigned char *dstppV = dstpV - dst_pitchUV;
	unsigned char *dstppU = dstpU - dst_pitchUV;
	unsigned char *dstpppV = dstppV - dst_pitchUV2;
	unsigned char *dstpppU = dstppU - dst_pitchUV2;
	unsigned char *dstpnV = dstpV + dst_pitchUV;
	unsigned char *dstpnU = dstpU + dst_pitchUV;
	unsigned char *dstpnnV = dstpnV + dst_pitchUV2;
	unsigned char *dstpnnU = dstpnU + dst_pitchUV2;
	for (x=0; x<dst_widthUV; ++x)
	{
		dstpV[x] = (dstppV[x] + dstpnV[x] + 1) >> 1;
		dstpU[x] = (dstppU[x] + dstpnU[x] + 1) >> 1;
	}
	dstpppV = dstppV;
	dstppV = dstpnV;
	dstpV += dst_pitchUV2;
	dstpnV = dstpnnV;
	dstpnnV += dst_pitchUV2;
	dstpppU = dstppU;
	dstppU = dstpnU;
	dstpU += dst_pitchUV2;
	dstpnU = dstpnnU;
	dstpnnU += dst_pitchUV2;
	temp1 = dst_heightUV-4;
	for (y=3; y<temp1; y+=2)
	{
		for (x=0; x<dst_widthUV; ++x) 
		{
			dstpV[x] = (*this.*cPtr)(dstpppV[x],dstppV[x],dstpnV[x],dstpnnV[x]);
			dstpU[x] = (*this.*cPtr)(dstpppU[x],dstppU[x],dstpnU[x],dstpnnU[x]);
		}
		dstpppV = dstppV;
		dstppV = dstpnV;
		dstpV += dst_pitchUV2;
		dstpnV = dstpnnV;
		dstpnnV += dst_pitchUV2;
		dstpppU = dstppU;
		dstppU = dstpnU;
		dstpU += dst_pitchUV2;
		dstpnU = dstpnnU;
		dstpnnU += dst_pitchUV2;
	}
	for (x=0; x<dst_widthUV; ++x)
	{
		dstpV[x] = (dstppV[x] + dstpnV[x] + 1) >> 1;
		dstpU[x] = (dstppU[x] + dstpnU[x] + 1) >> 1;
	}
	dstpV += dst_pitchUV2;
	dstpU += dst_pitchUV2;
	for (x=0; x<dst_widthUV; ++x)
	{
		dstpV[x] = dstpnV[x];
		dstpU[x] = dstpnU[x];
	}
	return Dst;
}

bool FastEDIUpsizer::checkVariance(int pc[4])
{
	int sum = pc[0] + pc[1] + pc[2] + pc[3];
	int sumsq = pc[0]*pc[0] + pc[1]*pc[1] + pc[2]*pc[2] + pc[3]*pc[3];
	if ((sumsq<<2) - sum*sum <= mvar) return false;
	return true;
}

unsigned char FastEDIUpsizer::nEDI_8x8(int p[64], int n[64][4], int c)
{
	int i, j, k, final;
	double a[4], CtC[4][4], r[4], test, s;
	double CtCi[4][4] = { 1.0f,  0.f,  0.f,  0.f,
						   0.f, 1.0f,  0.f,  0.f,
						   0.f,  0.f, 1.0f,  0.f,
						   0.f,  0.f,  0.f, 1.0f  };
	for(i=0; i<4; ++i) 
	{
		for(j=i; j<4; ++j) 
		{
			s = 0;
			for(k=0; k<64; ++k) s +=  n[k][i] * n[k][j];
			CtC[i][j] = CtC[j][i] = s;
		}
	}
	if (!(gaussianInvert(CtCi, CtC))) return fallBack(p, c);
	if (threshold > 0)
	{
		test = conditionNumber(CtCi, CtC);
		if (test >= threshold) return fallBack(p, c);
	}
	for(i=0; i<4; ++i) 
	{
		s = 0;
		for(k=0; k<64; ++k) s += n[k][i] * p[k];
		r[i] = s;
	}
	for(i=0; i<4; ++i)
	{
		s = 0;
		for (k=0; k<4; ++k) s += CtCi[i][k] * r[k];
		a[i] = s;
	}
	final = (int)(p[15]*a[0] + p[26]*a[1] + p[37]*a[2] + p[48]*a[3] + 0.5f);
	if (constraint == 1 && (final>255 || final<0)) return fallBack(p, c);
	else if (constraint == 2)
	{
		int highN = max(max(p[15],p[26]),max(p[37],p[48])) + 5;
		int lowN = min(min(p[15],p[26]),min(p[37],p[48])) - 5;
		if (final>highN || final<lowN) return fallBack(p, c);
	}
	if (final>255) final = 255;
	else if (final<0) final = 0;
	return (unsigned char)final;
}

bool FastEDIUpsizer::gaussianInvert(double CtCi[4][4], double CtCin[4][4])
{
	int best, i, j, k;
	double temp, scale, CtC[4][4];
	memcpy(CtC, CtCin, (sizeof(double)<<4));
	for(i=0; i<4; ++i)
	{
		best = i;
		for(j=i+1; j<4; ++j)
			if (fabs(CtC[best][i]) < fabs(CtC[j][i])) best = j;
		if (best != i)
		{
			for(k=0; k<4; ++k)
			{
				temp = CtC[i][k];
				CtC[i][k] = CtC[best][k];
				CtC[best][k] = temp;
				temp = CtCi[i][k];
				CtCi[i][k] = CtCi[best][k];
				CtCi[best][k] = temp;
			}
		}
		if (CtC[i][i] == 0) 
			return false;
		else
		{
			temp = CtC[i][i];
			for (k=0; k<4; ++k)
			{
				if (k != i)
				{
					scale = -(CtC[k][i]/temp);
					for(j=0; j<4; ++j)
					{
						CtC[k][j] += (CtC[i][j] * scale);
						CtCi[k][j] += (CtCi[i][j] * scale);
					}
					CtC[k][i] = 0.0f;
				}
			}
		}
	}
    for(i=0; i<4; ++i)
	{
		if (CtC[i][i] != 1)
		{
			scale = 1.0f/CtC[i][i];
			for(k=0; k<4; ++k) CtCi[i][k] *= scale;
			CtC[i][i] = 1.0f;
		}
	}
	return true;
}

double FastEDIUpsizer::conditionNumber(double CtCi[4][4], double CtC[4][4])
{
	double highest1 = fabs(CtCi[0][0]) + fabs(CtCi[0][1]) + fabs(CtCi[0][2]) + fabs(CtCi[0][3]);
	double highest2 = fabs(CtC[0][0]) + fabs(CtC[0][1]) + fabs(CtC[0][2]) + fabs(CtC[0][3]), temp;
	for(int i=1; i<4; ++i)
	{
		temp = fabs(CtCi[i][0]) + fabs(CtCi[i][1]) + fabs(CtCi[i][2]) + fabs(CtCi[i][3]);
		if (temp > highest1) highest1 = temp;
		temp = fabs(CtC[i][0]) + fabs(CtC[i][1]) + fabs(CtC[i][2]) + fabs(CtC[i][3]);
		if (temp > highest2) highest2 = temp;
	}
	return (highest1 * highest2);
}

unsigned char FastEDIUpsizer::fallBack(int p[64], int type)
{
	if (type == 1)
	{
		unsigned char temp1 = (*this.*cPtr)((unsigned char)p[12],(unsigned char)p[13],(unsigned char)p[24],(unsigned char)p[25]);
		unsigned char temp2 = (*this.*cPtr)((unsigned char)p[14],(unsigned char)p[15],(unsigned char)p[26],(unsigned char)p[27]);
		unsigned char temp3 = (*this.*cPtr)((unsigned char)p[36],(unsigned char)p[37],(unsigned char)p[48],(unsigned char)p[49]);
		unsigned char temp4 = (*this.*cPtr)((unsigned char)p[38],(unsigned char)p[39],(unsigned char)p[50],(unsigned char)p[51]);
		return (*this.*cPtr)(temp1,temp2,temp3,temp4);
	}
	else if (type == 2)
		return (*this.*cPtr)((unsigned char)p[12],(unsigned char)p[15],(unsigned char)p[48],(unsigned char)p[51]);
	else
		return (*this.*cPtr)((unsigned char)p[25],(unsigned char)p[26],(unsigned char)p[37],(unsigned char)p[38]);
}

unsigned char FastEDIUpsizer::cubic50(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4)
{
	int temp = (int)((9*(p2+p3)-p1-p4+8)>>4);
	if (temp>255) temp = 255; else if (temp<0) temp = 0;
	return (unsigned char)temp;
}

unsigned char FastEDIUpsizer::cubic625(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4)
{
	int temp = (int)((37*(p2+p3)-5*(p1+p4)+32)>>6);
	if (temp>255) temp = 255; else if (temp<0) temp = 0;
	return (unsigned char)temp;
}

unsigned char FastEDIUpsizer::cubic75(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4)
{
	int temp = (int)((19*(p2+p3)-3*(p1+p4)+16)>>5);
	if (temp>255) temp = 255; else if (temp<0) temp = 0;
	return (unsigned char)temp;
}

unsigned char FastEDIUpsizer::cubic875(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4)
{
	int temp = (int)((39*(p2+p3)-7*(p1+p4)+32)>>6);
	if (temp>255) temp = 255; else if (temp<0) temp = 0;
	return (unsigned char)temp;
}

unsigned char FastEDIUpsizer::cubic10(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4)
{
	int temp = (int)((5*(p2+p3)-p1-p4+4)>>3);
	if (temp>255) temp = 255; else if (temp<0) temp = 0;
	return (unsigned char)temp;
}

AVSValue __cdecl Create_FastEDIUpsizer(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	double threshold = 5000.0f;
	int constraint = 2;
	int cubic = 2;
	int mvar = 8;
    return new FastEDIUpsizer(
		args[0].AsClip(),
		args[1].AsFloat(threshold),
		args[2].AsInt(constraint),
		args[3].AsInt(cubic),
		args[4].AsInt(mvar),
		env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("FastEDIUpsizer", "c[threshold]f[constraint]i[cubic]i[mvar]i", Create_FastEDIUpsizer, 0);
    return 0;
}