/*
	FastEDIUpsizer plugin for Avisynth v2.5.x.  FastEDIUpsizer upscales video
	by a factor of two using NEDI with a constant 8x8 window size on luma and
	bicubic interpolation on the chroma planes.  Only works in YV12 colorspace.

	Copyright (C) 2004 Kevin Stone

	v0.9 - header file

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <math.h>
#include "internal.h"

class FastEDIUpsizer : public GenericVideoFilter
{	
	double threshold;	// matrix condition number threshold
	int constraint;		// extra constraint for edi
	int cubic;			// 'A' value for cubic interpolation
	int mvar;			// minimum local variance to use NEDI
	int nfrms;
	int pc[4], p[64], n[64][4], localPat1[4], localPat2[4];
	int windowPat161[16], windowPat162[16], windowPat641[64], windowPat642[64];
	unsigned char FastEDIUpsizer::nEDI_8x8(int p[64], int n[64][4], int c);
	unsigned char FastEDIUpsizer::fallBack(int p[64], int type);
	bool FastEDIUpsizer::gaussianInvert(double CtCi[4][4], double CtCin[4][4]);
	double FastEDIUpsizer::conditionNumber(double CtCi[4][4], double CtC[4][4]);
	bool FastEDIUpsizer::checkVariance(int pc[4]);
	unsigned char FastEDIUpsizer::cubic50(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4);
	unsigned char FastEDIUpsizer::cubic625(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4);
	unsigned char FastEDIUpsizer::cubic75(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4);
	unsigned char FastEDIUpsizer::cubic875(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4);
	unsigned char FastEDIUpsizer::cubic10(unsigned char p1, unsigned char p2, unsigned char p3, unsigned char p4);
	unsigned char (FastEDIUpsizer::*cPtr)(unsigned char, unsigned char, unsigned char, unsigned char);

public:
	PVideoFrame __stdcall GetFrame(int frameNum, IScriptEnvironment* env);

    FastEDIUpsizer(PClip _child, double _threshold, int _constraint, int _cubic, int _mvar,
		IScriptEnvironment* env) : GenericVideoFilter(_child), threshold(_threshold), 
		constraint(_constraint), cubic(_cubic), mvar(_mvar)
	{
		if (!vi.IsYV12()) 
			env->ThrowError("FastEDIUpsizer:  YV12 input only!");
		if (vi.height < 10)
			env->ThrowError("FastEDIUpsizer:  input height must be at least 10!");
		if ((vi.height&1) || (vi.width&1))
			env->ThrowError("FastEDIUpsizer:  input height and width must be multiples of 2!");
		if (constraint<0 || constraint>2)
			env->ThrowError("FastEDIUpsizer:  constraint must be set to 0, 1, or 2!");
		if (cubic < 0 || cubic > 4)
			env->ThrowError("FastEDIUpsizer:  cubic must be set to 0, 1, 2, 3, or 4!");
		vi.width <<= 1;
		vi.height <<= 1;
		nfrms = vi.num_frames - 1;
		mvar <<= 4;
		if (cubic == 0) cPtr = FastEDIUpsizer::cubic50;
		else if (cubic == 1) cPtr = FastEDIUpsizer::cubic625;
		else if (cubic == 2) cPtr = FastEDIUpsizer::cubic75;
		else if (cubic == 3) cPtr = FastEDIUpsizer::cubic875;
		else cPtr = FastEDIUpsizer::cubic10;
	}
	FastEDIUpsizer::~FastEDIUpsizer()
	{
		// nothing to free
	}
};